import itertools
import os

import pandas as pd
import torch
import torch.distributed as dist
import torch.multiprocessing as mp
from peft import PeftModel
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.utils.data import DataLoader, DistributedSampler, TensorDataset
from tqdm import tqdm
from transformers import AutoModelForSequenceClassification, AutoTokenizer, logging

logging.set_verbosity_error()


def read_fa(file_path):
    chrom_dic = {}
    with open(file_path, "r") as f:
        lines = pd.DataFrame(f.readlines(), columns=["sequence"])
        lines["sequence"] = lines["sequence"].str.rstrip("\n")

        mask = lines["sequence"].str.startswith(">")
        index = lines[mask].index
        for i in range(len(index)):
            chrom = lines.loc[index[i], "sequence"].lstrip(">").split(" ")[0]
            left = index[i] + 1
            right = index[i + 1] if i + 1 < len(index) else None
            chrom_dic["chr{}".format(chrom)] = "".join(lines[left:right]["sequence"])
            print("chr{}".format(chrom))

    return chrom_dic


def divide(datas: pd.DataFrame, valtest_percentage: float, random_seed: int):
    _train = datas.sample(frac=1 - valtest_percentage, random_state=random_seed)
    _val = datas.drop(_train.index)
    return _train, _val


def _get_model_and_dataset(
    model_name_or_path: str,
    lora_model_path: str,
    model_max_length: int,
    texts: list,
):
    tokenizer = AutoTokenizer.from_pretrained(
        model_name_or_path,
        model_max_length=model_max_length,
        padding_side="right",
        use_fast=True,
        trust_remote_code=True,
    )
    model = AutoModelForSequenceClassification.from_pretrained(
        model_name_or_path, num_labels=1, trust_remote_code=True
    )
    if lora_model_path:
        model = PeftModel.from_pretrained(model, lora_model_path)

    input_ids = tokenizer(
        texts,
        return_tensors="pt",
        padding="max_length",
        max_length=tokenizer.model_max_length,
        truncation=True,
    )["input_ids"]
    dataset = TensorDataset(torch.arange(input_ids.shape[0]), input_ids)

    return model, dataset


def classify(
    texts: list,
    model_name_or_path: str = "liminghong/DNABERT-2-117M",
    lora_model_path: str = os.path.join(os.path.dirname(__file__), "adapter_model"),
    model_max_length: int = 1250,
    batch_size: int = 32,
):
    model, dataset = _get_model_and_dataset(
        model_name_or_path, lora_model_path, model_max_length, texts
    )

    model.cuda()
    model.eval()

    results = []
    for _, input_ids in tqdm(DataLoader(dataset, batch_size=batch_size, shuffle=False)):
        with torch.no_grad():
            _output = model(input_ids.cuda())
        results.extend(_output["logits"].flatten().tolist())
    return results


def _ddp_inference(rank, world_size, queue, model, dataset, batch_size):
    dist.init_process_group(backend="gloo", rank=rank, world_size=world_size)
    torch.cuda.set_device(rank)

    model = model.to(rank)
    model = DDP(model, device_ids=[rank])
    model.eval()

    sampler = DistributedSampler(
        dataset, num_replicas=world_size, rank=rank, shuffle=False
    )
    data_loader = DataLoader(
        dataset, sampler=sampler, batch_size=batch_size, shuffle=False
    )

    results = []
    for index, input_ids in data_loader:
        with torch.no_grad():
            _output = model(input_ids.to(rank))
        results.extend(list(zip(index.tolist(), _output["logits"].flatten().tolist())))
    queue.put(results)

    dist.destroy_process_group()


def parallel_classify(
    texts: list,
    model_name_or_path: str = "liminghong/DNABERT-2-117M",
    lora_model_path: str = os.path.join(os.path.dirname(__file__), "adapter_model"),
    model_max_length: int = 1250,
    world_size: int = torch.cuda.device_count(),
    batch_size: int = 32,
    MASTER_ADDR: str = "localhost",
    MASTER_PORT: str = "12355",
):
    model, dataset = _get_model_and_dataset(
        model_name_or_path, lora_model_path, model_max_length, texts
    )

    os.environ["TOKENIZERS_PARALLELISM"] = "false"
    os.environ["MASTER_ADDR"] = MASTER_ADDR
    os.environ["MASTER_PORT"] = MASTER_PORT

    ctx = mp.get_context("spawn")
    queue = ctx.Queue()
    for i in range(world_size):
        p = ctx.Process(
            target=_ddp_inference,
            args=(i, world_size, queue, model, dataset, batch_size),
        )
        p.start()
    results = [queue.get() for _ in range(world_size)]

    results = list(itertools.chain(*results))
    results = (
        pd.DataFrame(results).drop_duplicates().sort_values(by=0)[1].values.tolist()
    )

    return results



# -*- coding: utf-8 -*-
# Copyright (c) 2021, Tencent Inc. All rights reserved.
# Aucthor: chenchenqin
# Data: 2021/3/8
import os
import random
from datetime import datetime

import numpy as np
import torch
import torch.nn.functional as F

__all__ = ["seed_all_rng"]


def top_k_top_p_filter(logits, top_k: int = 0, top_p: float = 0.0):
    if top_k > 0:
        filter, _ = torch.topk(logits, min(top_k, logits.size(-1)))
        logits[logits < filter[:, [-1]]] = float('-inf')

    if top_p > 0.0:
        sorted_logits, sorted_indices = torch.sort(logits, descending=True)
        cumulative_probs = torch.cumsum(
            F.softmax(sorted_logits, dim=-1), dim=-1)
        filter = cumulative_probs > top_p
        filter[..., 1:] = filter[..., :-1].clone()
        filter[..., 0] = 0
        indices_to_remove = filter.scatter(1, sorted_indices, filter)
        logits[indices_to_remove] = float('-inf')

    return logits


def seed_all_rng(seed=None):
    """
    Set the random seed for the RNG in torch, numpy and python.

    Args:
        seed (int): if None, will use a strong random seed.
    """
    if seed is None:
        seed = (os.getpid() +
                int(datetime.now().strftime("%S%f")) +
                int.from_bytes(os.urandom(2), "big")
                )
    np.random.seed(seed)
    torch.set_rng_state(torch.manual_seed(seed).get_state())
    random.seed(seed)



# -*- coding: utf-8 -*-
# Copyright (c) 2023, Tencent Inc. All rights reserved.
# Author: chenchenqin
# Data: 2023/9/4 12:01
import argparse
import os

import torch

from dna_gpt.model import DNAGPT
from dna_gpt.tokenizer import KmerTokenizer
from dna_gpt.utils import seed_all_rng


def get_model(model_name):
    special_tokens = (['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'] +
                      ["+", '-', '*', '/', '=', "&", "|", "!"] +
                      ['M', 'B'] + ['P'] + ['R', 'I', 'K', 'L', 'O', 'Q', 'S', 'U', 'V'] + ['W', 'Y', 'X', 'Z'])
    if model_name in ('dna_gpt0.1b_h',):
        tokenizer = KmerTokenizer(6, special_tokens, dynamic_kmer=False)
    else:
        tokenizer = KmerTokenizer(6, special_tokens, dynamic_kmer=True)

    vocab_size = len(tokenizer)
    model = DNAGPT.from_name(model_name, vocab_size)
    return model, tokenizer


def load_model(model, weight_path, device=None, dtype=None):
    state = torch.load(weight_path, map_location="cpu")
    if 'model' in state.keys():
        model.load_state_dict(state['model'], strict=False)
    else:
        model.load_state_dict(state, strict=False)
    print(f"loading model weights from {weight_path}")
    model.to(device=device, dtype=dtype)
    model = model.eval()
    return model


def generate(model,
             tokenizer,
             prompt,
             max_len=256,
             num_samples=1,
             temperature=1.0,
             top_k=0,
             top_p=0):
    print(f"max length is {max_len}")
    device = next(model.parameters()).device
    prompt_ids = tokenizer.encode(prompt, max_len=max_len, device=device)
    print(f"prompt token ids: {prompt_ids.tolist()}")
    max_new_tokens = max_len - len(prompt_ids)
    for k in range(num_samples):
        x = prompt_ids[None, :]
        y = model.generate(x,
                           max_new_tokens,
                           temperature=temperature,
                           top_k=top_k,
                           top_p=top_p,
                           do_sample=True,
                           stop_ids=(tokenizer.unk_id, tokenizer.pad_id))
        output = tokenizer.decode(y[0].tolist())
        output = output[len(prompt):]
        print(f"\033[31m{prompt}\033[0m{output}")
        print('-' * 50)


def regression(model, tokenizer, prompt, numbers, max_len=256):
    device = next(model.parameters()).device
    dtype = next(model.parameters()).dtype
    prompt_ids = tokenizer.encode(prompt, max_len=max_len, device=device)
    prompt_len = len(prompt_ids)
    x = prompt_ids[None]
    num_len = len(numbers)
    print(f"inputs numbers: {numbers}")
    token_length = torch.tensor(prompt_len + num_len - 1, device=device, dtype=torch.long)[None, None]
    number_block = torch.full((x.shape[0],), x.shape[1] - 2)
    num = torch.tensor(numbers, device=device, dtype=dtype)
    y = model(x, num, token_length, number_block=number_block)
    print("The predicted expression level is:", f"\033[0;31m{y[0][0, 0, 0].tolist()}\033[0m")


def classification(model, tokenizer, prompt, max_len=256):
    device = next(model.parameters()).device
    prompt_ids = tokenizer.encode(prompt, max_len=max_len, device=device)[None]
    output = model(prompt_ids)[:, -1]
    y = torch.argmax(output, dim=-1)
    output = tokenizer.decode([y[0]])
    if output == 'N':
        print("\033[0;32mThe input is a real GSR sequence!\033[0m")
    elif output == 'A':
        print("\033[0;31mThe input is a fake GSR sequence!\033[0m")


def main(args):
    torch.set_grad_enabled(False)
    seed_all_rng(args.seed)
    model_name = args.name
    weight_path = args.weight or f"checkpoints/{model_name}.pth"
    assert os.path.exists(weight_path), f"not exist checkpoint: {weight_path}"
    prompt = args.input
    device = args.device or ('cuda' if torch.cuda.is_available() else 'cpu')
    dtype = args.dtype or 'float16'
    dtype = getattr(torch, dtype)
    model, tokenizer = get_model(model_name)
    model = load_model(model, weight_path, device=device, dtype=dtype)
    max_len = min(args.max_len, model.max_len)
    if args.task == "generation":
        generate(model, tokenizer, prompt,
                 max_len=max_len,
                 num_samples=args.num_samples,
                 temperature=args.temperature,
                 top_k=args.topk,
                 top_p=args.topp)
    elif args.task == "regression":
        numbers = [float(n) for n in args.numbers.split()]
        regression(model, tokenizer, prompt, numbers, max_len=max_len)
    elif args.task == "classification":
        classification(model, tokenizer, prompt, max_len=max_len)
    else:
        raise f"no task type: {args.task}"


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='sample dna gpt model generation')
    parser.add_argument('--task', default="generation", help='dtype of the model weights')
    parser.add_argument('--input', '-i', default='<R>CTGTATACCACAGA', help='input prompt text')
    parser.add_argument('--numbers', help='input number list')
    parser.add_argument('--name', '-n', default='dna_gpt0.1b_m', help='type of the model')
    parser.add_argument('--num_samples', '-ns', type=int, default=10, help='number samples of generation')
    parser.add_argument('--max_len', '-ml', type=int, default=256, help='max length of input token ids')
    parser.add_argument('--temperature', type=float, default=1.0, help='sample temperature')
    parser.add_argument('--topk', type=int, default=0, help='sample topk')
    parser.add_argument('--topp', type=float, default=0.95, help='sample topp')
    parser.add_argument('--seed', default=40, type=int, help='random seed for sampling')
    parser.add_argument('--weight', '-w', default=None, help='path to the weight')
    parser.add_argument('--device', default=None, help='device of the model, cuda or cpu')
    parser.add_argument('--dtype', default=None, help='dtype of the model weights')
    args = parser.parse_args()
    main(args)


import json
import logging
import os
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Sequence

import pandas as pd
import torch
import transformers
from peft import LoraConfig, get_peft_model
from torch.utils.data import Dataset
from transformers import TrainerControl, TrainerState
from transformers.integrations import NeptuneCallback
from transformers.trainer_callback import TrainerCallback
from transformers.trainer_utils import PREFIX_CHECKPOINT_DIR


@dataclass
class ModelArguments:
    model_name_or_path: Optional[str] = field(default="facebook/opt-125m")
    use_lora: bool = field(default=False, metadata={"help": "whether to use LoRA"})
    lora_r: int = field(default=8, metadata={"help": "hidden dimension for LoRA"})
    lora_alpha: int = field(default=32, metadata={"help": "alpha for LoRA"})
    lora_dropout: float = field(
        default=0.05, metadata={"help": "dropout rate for LoRA"}
    )
    lora_target_modules: str = field(
        default="query,value", metadata={"help": "where to perform LoRA"}
    )
    modules_to_save: str = field(default="", metadata={"help": "where to perform LoRA"})


@dataclass
class DataArguments:
    data_path: str = field(
        default=None, metadata={"help": "Path to the training data."}
    )
    kmer: int = field(
        default=-1,
        metadata={"help": "k-mer for input sequence. -1 means not using k-mer."},
    )


@dataclass
class TrainingArguments(transformers.TrainingArguments):
    cache_dir: Optional[str] = field(default=None)
    run_name: str = field(default="run")
    optim: str = field(default="adamw_torch")
    model_max_length: int = field(
        default=512, metadata={"help": "Maximum sequence length."}
    )
    gradient_accumulation_steps: int = field(default=1)
    per_device_train_batch_size: int = field(default=1)
    per_device_eval_batch_size: int = field(default=1)
    num_train_epochs: int = field(default=1)
    fp16: bool = field(default=False)
    logging_steps: int = field(default=100)
    save_steps: int = field(default=100)
    eval_steps: int = field(default=100)
    evaluation_strategy: str = (field(default="steps"),)
    lr_scheduler_type: str = field(default="linear")
    warmup_steps: int = field(default=50)
    weight_decay: float = field(default=0.01)
    learning_rate: float = field(default=1e-4)
    save_total_limit: int = field(default=None)
    load_best_model_at_end: bool = field(default=True)
    output_dir: str = field(default="output")
    find_unused_parameters: bool = field(default=False)
    ddp_find_unused_parameters: bool = field(default=False)
    checkpointing: bool = field(default=False)
    dataloader_pin_memory: bool = field(default=False)
    eval_and_save_results: bool = field(default=True)
    save_model: bool = field(default=False)
    seed: int = field(default=42)
    report_to: str = field(default="none")


def safe_save_model_for_hf_trainer(trainer: transformers.Trainer, output_dir: str):
    """Collects the state dict and dump to disk."""
    state_dict = trainer.model.state_dict()
    if trainer.args.should_save:
        cpu_state_dict = {key: value.cpu() for key, value in state_dict.items()}
        del state_dict
        trainer._save(output_dir, state_dict=cpu_state_dict)  # noqa


"""
Get the reversed complement of the original DNA sequence.
"""


def get_alter_of_dna_sequence(sequence: str):
    MAP = {"A": "T", "T": "A", "C": "G", "G": "C"}
    # return "".join([MAP[c] for c in reversed(sequence)])
    return "".join([MAP[c] for c in sequence])


"""
Transform a dna sequence to k-mer string
"""


def generate_kmer_str(sequence: str, k: int) -> str:
    """Generate k-mer string from DNA sequence."""
    return " ".join([sequence[i : i + k] for i in range(len(sequence) - k + 1)])


"""
Load or generate k-mer string for each DNA sequence. The generated k-mer string will be saved to the same directory as the original data with the same name but with a suffix of "_{k}mer".
"""


def load_or_generate_kmer(data_path: str, texts: List[str], k: int) -> List[str]:
    """Load or generate k-mer string for each DNA sequence."""
    kmer_path = data_path.replace(".csv", f"_{k}mer.json")
    if os.path.exists(kmer_path):
        logging.warning(f"Loading k-mer from {kmer_path}...")
        with open(kmer_path, "r") as f:
            kmer = json.load(f)
    else:
        logging.warning(f"Generating k-mer...")
        kmer = [generate_kmer_str(text, k) for text in texts]
        with open(kmer_path, "w") as f:
            logging.warning(f"Saving k-mer to {kmer_path}...")
            json.dump(kmer, f)

    return kmer


class SupervisedDataset(Dataset):
    """Dataset for supervised fine-tuning."""

    def __init__(
        self,
        data_path: str,
        tokenizer: transformers.PreTrainedTokenizer,
        kmer: int = -1,
    ):
        super(SupervisedDataset, self).__init__()

        # load data from the disk
        data = pd.read_csv(data_path, index_col=0)
        texts = data["sequence"].values.tolist()
        labels = data["label"].values.tolist()

        if kmer != -1:
            # only write file on the first process
            if torch.distributed.get_rank() not in [0, -1]:
                torch.distributed.barrier()

            logging.warning(f"Using {kmer}-mer as input...")
            texts = load_or_generate_kmer(data_path, texts, kmer)

            if torch.distributed.get_rank() == 0:
                torch.distributed.barrier()

        self.texts = texts
        self.labels = labels
        self.tokenizer = tokenizer
        self.num_labels = 1

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, i) -> Dict[str, torch.Tensor]:
        return dict(
            input_ids=self.tokenizer(
                self.texts[i],
                return_tensors="pt",
                padding="max_length",
                max_length=self.tokenizer.model_max_length,
                truncation=True,
            )["input_ids"][0],
            labels=self.labels[i],
        )


@dataclass
class DataCollatorForSupervisedDataset(object):
    """Collate examples for supervised fine-tuning."""

    tokenizer: transformers.PreTrainedTokenizer

    def __call__(self, instances: Sequence[Dict]) -> Dict[str, torch.Tensor]:
        input_ids, labels = tuple(
            [instance[key] for instance in instances] for key in ("input_ids", "labels")
        )
        input_ids = torch.nn.utils.rnn.pad_sequence(
            input_ids, batch_first=True, padding_value=self.tokenizer.pad_token_id
        )
        labels = torch.Tensor(labels)
        return dict(
            input_ids=input_ids,
            labels=labels,
            attention_mask=input_ids.ne(self.tokenizer.pad_token_id),
        )


class SavePeftModelCallback(TrainerCallback):
    def on_save(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        checkpoint_folder = os.path.join(
            args.output_dir, f"{PREFIX_CHECKPOINT_DIR}-{state.global_step}"
        )

        peft_model_path = os.path.join(checkpoint_folder, "adapter_model")
        kwargs["model"].save_pretrained(peft_model_path)

        pytorch_model_path = os.path.join(checkpoint_folder, "pytorch_model.bin")
        try:
            os.remove(pytorch_model_path)
        except:
            pass

        return control


def train():
    parser = transformers.HfArgumentParser(
        (ModelArguments, DataArguments, TrainingArguments)
    )
    model_args, data_args, training_args = parser.parse_args_into_dataclasses()

    # load tokenizer
    tokenizer = transformers.AutoTokenizer.from_pretrained(
        model_args.model_name_or_path,
        cache_dir=training_args.cache_dir,
        model_max_length=training_args.model_max_length,
        padding_side="right",
        use_fast=True,
        trust_remote_code=True,
    )

    if "InstaDeepAI" in model_args.model_name_or_path:
        tokenizer.eos_token = tokenizer.pad_token

    # define datasets and data collator
    train_dataset = SupervisedDataset(
        tokenizer=tokenizer,
        data_path=os.path.join(data_args.data_path, "train.csv"),
        kmer=data_args.kmer,
    )
    val_dataset = SupervisedDataset(
        tokenizer=tokenizer,
        data_path=os.path.join(data_args.data_path, "dev.csv"),
        kmer=data_args.kmer,
    )
    data_collator = DataCollatorForSupervisedDataset(tokenizer=tokenizer)

    # load model
    model = transformers.AutoModelForSequenceClassification.from_pretrained(
        model_args.model_name_or_path,
        cache_dir=training_args.cache_dir,
        num_labels=train_dataset.num_labels,
        trust_remote_code=True,
        output_hidden_states=False,
    )

    # configure LoRA
    if model_args.use_lora:
        lora_config = LoraConfig(
            r=model_args.lora_r,
            lora_alpha=model_args.lora_alpha,
            target_modules=list(model_args.lora_target_modules.split(",")),
            lora_dropout=model_args.lora_dropout,
            bias="none",
            modules_to_save=list(model_args.modules_to_save.split(",")),
            task_type="SEQ_CLS",
            inference_mode=False,
        )
        model = get_peft_model(model, lora_config)
        model.print_trainable_parameters()

    # define trainer
    trainer = transformers.Trainer(
        model=model,
        tokenizer=tokenizer,
        args=training_args,
        train_dataset=train_dataset,
        eval_dataset=val_dataset,
        data_collator=data_collator,
        callbacks=[
            SavePeftModelCallback,  # if lora
            # NeptuneCallback(log_parameters=False),  # if report to neptune
        ],
    )
    trainer.train(ignore_keys_for_eval=["hidden_states"])

    if training_args.save_model:
        trainer.save_state()
        safe_save_model_for_hf_trainer(
            trainer=trainer, output_dir=training_args.output_dir
        )

    # get the evaluation results from trainer
    if training_args.eval_and_save_results and not model_args.use_lora:
        results_path = os.path.join(
            training_args.output_dir, "results", training_args.run_name
        )
        results = trainer.evaluate(
            eval_dataset=val_dataset, ignore_keys=["hidden_states"]
        )
        os.makedirs(results_path, exist_ok=True)
        with open(os.path.join(results_path, "eval_results.json"), "w") as f:
            json.dump(results, f)


if __name__ == "__main__":
    train()



import os

import click
import numpy as np
import pandas as pd
import pyBigWig
from scipy.signal import argrelextrema
from scipy.stats import gaussian_kde
from sklearn.preprocessing import QuantileTransformer

from bioseba.util import divide, read_fa


@click.group()
def cli():
    pass


@cli.command()
@click.option(
    "--data_dir_path", help="Folder path for insulation score files (.bw)", type=str
)
@click.option("--fasta_path", type=str)
@click.option("--output_path", type=str)
def prepross(data_dir_path, fasta_path, output_path):
    file_paths = [
        os.path.join(data_dir_path, file)
        for _, _, files in os.walk(data_dir_path, topdown=False)
        for file in files
    ]

    insulation_scores = None
    for file_path in file_paths:
        bw = pyBigWig.open(file_path)
        insulation_score = pd.DataFrame()
        for chrom in bw.chroms():
            _insulation_score = pd.DataFrame(
                bw.intervals(chrom),
                columns=[
                    "start",
                    "end",
                    os.path.splitext(os.path.basename(file_path))[0],
                ],
            )
            _insulation_score["chrom"] = chrom
            insulation_score = pd.concat([insulation_score, _insulation_score])
        insulation_score = insulation_score.set_index(["chrom", "start", "end"])

        if insulation_scores is None:
            insulation_scores = insulation_score
        else:
            insulation_scores = pd.concat([insulation_scores, insulation_score], axis=1)
    insulation_scores = insulation_scores.sort_index()

    cdbs = insulation_scores.copy()
    fa_dic = read_fa(fasta_path)

    def _generate_sequence(series):
        chrom, start, end = series.name
        return fa_dic[chrom][start:end]

    cdbs["sequence"] = cdbs.apply(_generate_sequence, axis=1)

    cdbs.to_csv(output_path, sep="\t")


@cli.command()
@click.option("--file_path", type=str, help="The path of the file to be processed")
@click.option("--output_path", type=str)
@click.option(
    "--valtest_percentage",
    default=0.2,
    type=float,
    help="Percentage of data used for val",
)
@click.option("--random_seed", default=42, type=int)
def create_dataset(file_path, output_path, valtest_percentage, random_seed):
    raw_datas = pd.read_csv(file_path, sep="\t")

    for chrom in raw_datas["chrom"].unique():
        for column in raw_datas.columns.drop(["chrom", "start", "end", "sequence"]):
            q1 = raw_datas.loc[raw_datas["chrom"] == chrom, column].quantile(0.25)
            _data = raw_datas.loc[
                (raw_datas["chrom"] == chrom) & (raw_datas[column] <= q1),
                ["start", column],
            ].copy()
            local_min_indices = argrelextrema(_data[column].values, np.less)[0]
            raw_datas.loc[
                (raw_datas["chrom"] == chrom)
                & (raw_datas["start"].isin(_data["start"].iloc[local_min_indices])),
                column,
            ] = 1
            raw_datas.loc[
                (raw_datas["chrom"] == chrom)
                & (~raw_datas["start"].isin(_data["start"].iloc[local_min_indices])),
                column,
            ] = 0
    raw_datas["SCI-CDB"] = raw_datas.drop(
        columns=["chrom", "start", "end", "sequence"]
    ).sum(axis=1)

    datas = pd.DataFrame()
    for chrom in raw_datas["chrom"].unique():
        _data = raw_datas.loc[raw_datas["chrom"] == chrom, ["start", "SCI-CDB"]].copy()
        _kde = gaussian_kde(_data["SCI-CDB"], bw_method=5)
        _peaks = _kde.pdf(_data["SCI-CDB"])
        local_max_indices = argrelextrema(_peaks, np.less)[0]
        _data = raw_datas.loc[
            (raw_datas["chrom"] == chrom)
            & (raw_datas["start"].isin(_data["start"].iloc[local_max_indices]))
        ].copy()
        datas = pd.concat([datas, _data])

    datas = datas[~datas["sequence"].str.contains("N")]

    quantile_transformer = QuantileTransformer(
        output_distribution="normal", random_state=random_seed
    )
    datas["label"] = quantile_transformer.fit_transform(
        np.expand_dims(datas["SCI-CDB"].values, axis=-1)
    ).flatten()
    datas = datas[datas["label"] > -4]

    train, val = divide(datas, valtest_percentage, random_seed)
    train.to_csv(os.path.join(output_path, "train.csv"))
    val.to_csv(os.path.join(output_path, "dev.csv"))


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
# Copyright (c) 2022, Tencent Inc. All rights reserved.
# Author: chenchenqin
# Data: 2022/12/14 19:15
import torch
import torch.nn as nn

from .gpt import GPT, LayerNorm


class DNAGPT(GPT):
    """ DNAGPT gene sequence model

    References:
        1) the official GPT-2 TensorFlow implementation released by OpenAI:
        https://github.com/openai/gpt-2/blob/master/src/model.py
        2) huggingface/transformers PyTorch implementation:
        https://github.com/huggingface/transformers/blob/main/src/transformers/models/gpt2/modeling_gpt2.py
    """

    def __init__(self,
                 vocab_size=7,
                 max_len=1024,
                 num_layers=3,
                 num_heads=3,
                 embedding_dim=48,
                 bias=True
                 ):
        super().__init__(vocab_size,
                         max_len,
                         num_layers,
                         num_heads,
                         embedding_dim,
                         bias=bias,
                         include_head=False)
        self.number_embedding = nn.Sequential(
            nn.Linear(1, self.embedding_dim, bias=bias),
            nn.SiLU(inplace=True),
            LayerNorm(self.embedding_dim, bias=bias),
            nn.Linear(self.embedding_dim, self.embedding_dim, bias=bias)
        )
        self.mlm_head = nn.Sequential(
            nn.Linear(self.embedding_dim, self.embedding_dim, bias=bias),
            nn.SiLU(inplace=True),
            LayerNorm(self.embedding_dim, bias=bias),
            nn.Linear(self.embedding_dim, vocab_size, bias=bias)
        )
        self.num_regression = nn.Sequential(
            nn.Linear(self.embedding_dim, self.embedding_dim, bias=bias),
            nn.SiLU(inplace=True),
            LayerNorm(self.embedding_dim, bias=bias),
            nn.Linear(self.embedding_dim, 1, bias=bias)
        )

    def _embedding_impl(self,
                        tokens,
                        numbers=None,
                        number_block=None):
        bs, _ = tokens.size()
        token_emb = self.transformer.wte(tokens)
        if numbers is not None and number_block is not None:
            numbers = numbers.to(token_emb.dtype)
            num_emb = self.number_embedding(numbers.unsqueeze(-1))
            output_tokens = []
            for j in range(bs):
                split_token = torch.split(token_emb[j], number_block[j].tolist(), dim=0)
                num_splits = len(split_token)
                stored_tokens = split_token[0]
                for i in range(num_splits - 1):
                    stored_tokens = torch.cat((stored_tokens, num_emb.unsqueeze(0)[j], split_token[i + 1]), dim=0)

                output_tokens.append(stored_tokens)

            token_emb = torch.stack(output_tokens, dim=0)

        seq_len = token_emb.shape[1]
        pos = torch.arange(0, seq_len,
                           dtype=torch.long,
                           device=token_emb.device).unsqueeze(0)
        pos_emb = self.transformer.wpe(pos)
        emb = token_emb + pos_emb
        return emb

    def _transformer_impl(self, embeddings):
        x = self.transformer.drop(embeddings)
        for block in self.transformer.h:
            x = block(x)

        return self.transformer.ln_f(x)

    def _head_impl(self, hiddens, number_loc=None):
        mlm = self.mlm_head(hiddens)
        if number_loc is None:
            return mlm

        bs, seq_len = hiddens.shape[:2]
        total_num = []
        for i in range(bs):
            bs_num = []
            for j in range(number_loc.shape[1]):
                bs_num.append(hiddens[i, number_loc[i, j]])
            bs_num = torch.stack(bs_num, dim=0)
            total_num.append(bs_num)
        total_num = torch.stack(total_num, dim=0)  # [bs, ]
        num = self.num_regression(total_num)
        return num, mlm

    def forward(self,
                token_ids,
                numbers=None,
                number_loc=None,
                number_block=None):
        x = self._embedding_impl(token_ids, numbers, number_block)
        x = self._transformer_impl(x)
        return self._head_impl(x, number_loc)

    @classmethod
    def from_name(cls, name, vocab_size):
        model_cfgs = {
            'dna_gpt0.1b_h': dict(vocab_size=vocab_size,
                                  max_len=4096,
                                  num_layers=12,
                                  num_heads=12,
                                  embedding_dim=768,
                                  bias=False),
            'dna_gpt0.1b_m': dict(vocab_size=vocab_size,
                                  max_len=512,
                                  num_layers=12,
                                  num_heads=12,
                                  embedding_dim=768,
                                  bias=False),
            'dna_gpt3b_m': dict(vocab_size=vocab_size,
                                max_len=512,
                                num_layers=60,
                                num_heads=64,
                                embedding_dim=2048,
                                bias=False)
        }
        assert name in model_cfgs, f"unkown model name, only suport: {list(model_cfgs.keys())}"
        cfg = model_cfgs[name]
        return cls(**cfg)


# -*- coding: utf-8 -*-
# Copyright (c) 2023, Tencent Inc. All rights reserved.
# Author: chenchenqin
# Data: 2023/9/4 12:01
import itertools as it
import re
from typing import Optional, Union

import numpy as np
import torch


class KmerTokenizer:

    def __init__(self,
                 k=3,
                 reserved_tokens=None,
                 dynamic_kmer=True):
        self.k = k
        if reserved_tokens is None:
            reserved_tokens = []
        assert len(reserved_tokens) == len(set(reserved_tokens)), "duplicated token in list"
        self.reserved_tokens = [f"<{t}>" for t in reserved_tokens]
        # N is unkown base in reference seqence
        self.bases = 'NAGCT'
        self.kmers = self.get_base_kmers(self.k, dynamic_kmer=dynamic_kmer)
        self.idx_to_token = self.reserved_tokens + self.kmers
        self.token_to_idx = {token: idx for idx, token in enumerate(self.idx_to_token)}
        self.pad_id = self.token_to_idx['<P>']
        self.unk_id = 0

    def get_base_kmers(self, k, dynamic_kmer=True):
        kmers = []
        start = 1 if dynamic_kmer else k
        for i in range(start, k + 1):
            kmers += list(it.product(self.bases, repeat=i))
        return [''.join(m) for m in kmers]

    def __len__(self):
        return len(self.idx_to_token)

    def piece_to_id(self, token: str):
        return self.token_to_idx.get(token, self.unk_id)

    def id_to_piece(self, token_id):
        assert 0 <= token_id < len(self), f"out of range of token id {token_id}, max {len(self)}"
        return self.idx_to_token[token_id]

    def _encode(self, token):
        if not isinstance(token, (list, tuple, np.ndarray)):
            return self.piece_to_id(token)
        return [self._encode(t) for t in token]

    def tokenize(self, text):
        tokens = re.split(r"[<>]", text)
        new_tokens = []
        n = self.k
        for t in tokens:
            if not t:
                continue

            if f"<{t}>" in self.reserved_tokens:
                new_tokens.append(f"<{t}>")
            else:
                seq = t
                # split kmers
                chunks = [seq[i:i + n] for i in range(0, len(seq), n)]
                new_tokens += chunks
        return new_tokens

    def encode(self,
               text,
               max_len: int = -1,
               pad: bool = False,
               device: Optional[torch.device] = None,
               to_tensor=True):
        pieces = self.tokenize(text)
        tokens = self._encode(pieces)
        if max_len > 0:
            tokens = tokens[:max_len]

        if pad and len(tokens) < max_len:
            tokens += [self.pad_id] * (max_len - len(tokens))

        if to_tensor:
            tokens = torch.tensor(tokens, dtype=torch.long, device=device)

        return tokens

    def decode(self, token_ids: Union[torch.Tensor, np.ndarray]) -> str:
        if isinstance(token_ids, (torch.Tensor, np.ndarray)):
            token_ids = token_ids.tolist()

        seq = "".join([self.id_to_piece(tid) for tid in token_ids])
        return seq


import os

import click
import gdown
import pandas as pd
import torch
from tqdm import tqdm

from bioseba.DNAGPT.dna_gpt.model import DNAGPT
from bioseba.DNAGPT.dna_gpt.tokenizer import KmerTokenizer
from bioseba.DNAGPT.dna_gpt.utils import seed_all_rng
from bioseba.util import classify, read_fa


@click.group()
def cli():
    pass


@cli.command()
@click.option("--data_dir_path", help="Folder path for boundary files", type=str)
@click.option("--fasta_path", type=str)
@click.option("--output_path", type=str)
def get_sequence_seed(data_dir_path, fasta_path, output_path):
    file_paths = [
        os.path.join(data_dir_path, file)
        for _, _, files in os.walk(data_dir_path, topdown=False)
        for file in files
    ]

    datas = pd.DataFrame()
    for file_path in tqdm(file_paths):
        _data = pd.read_csv(
            file_path,
            compression="gzip",
            sep="\t",
            header=None,
            names=["chrom", "start", "end", "level", "score"],
        )
        _data["source"] = os.path.splitext(os.path.basename(file_path))[0]
        datas = pd.concat([datas, _data])
    datas = datas.set_index(["chrom", "start", "end", "source"])

    fa_dic = read_fa(fasta_path)

    def _generate_target(series):
        chrom, start, end, source = series.name
        return fa_dic[chrom][start:end]

    datas["target"] = datas.apply(_generate_target, axis=1)

    def _generate_sequence(series):
        flank_len = 100
        chrom, start, end, source = series.name
        return fa_dic[chrom][start : start + flank_len]

    datas["sequence"] = datas.apply(_generate_sequence, axis=1)

    datas = datas[~datas["sequence"].str.contains("N")]
    datas = datas[~datas["target"].str.contains("N")]

    count = datas.groupby(["target"]).size().reset_index(name="counts")
    count = count[count["counts"] > 10]
    datas = pd.merge(count, datas, on=["target"], how="left").drop_duplicates(
        subset="target"
    )

    datas.to_csv(output_path, sep="\t")


def get_model(model_name):
    special_tokens = (
        ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
        + ["+", "-", "*", "/", "=", "&", "|", "!"]
        + ["M", "B"]
        + ["P"]
        + ["R", "I", "K", "L", "O", "Q", "S", "U", "V"]
        + ["W", "Y", "X", "Z"]
    )
    if model_name in ("dna_gpt0.1b_h",):
        tokenizer = KmerTokenizer(6, special_tokens, dynamic_kmer=False)
    else:
        tokenizer = KmerTokenizer(6, special_tokens, dynamic_kmer=True)

    vocab_size = len(tokenizer)
    model = DNAGPT.from_name(model_name, vocab_size)
    return model, tokenizer


def load_model(model, weight_path, device=None, dtype=None):
    state = torch.load(weight_path, map_location="cpu")
    if "model" in state.keys():
        model.load_state_dict(state["model"], strict=False)
    else:
        model.load_state_dict(state, strict=False)
    print(f"loading model weights from {weight_path}")
    model.to(device=device, dtype=dtype)
    model = model.eval()
    return model


def generate(
    model,
    tokenizer,
    prompt,
    max_len=256,
    num_samples=1,
    temperature=1.0,
    top_k=0,
    top_p=0,
):
    print(f"max length is {max_len}")
    device = next(model.parameters()).device
    prompt_ids = tokenizer.encode(prompt, max_len=max_len, device=device)
    print(f"prompt token ids: {prompt_ids.tolist()}")
    max_new_tokens = max_len - len(prompt_ids)
    outputs = []
    for k in tqdm(range(num_samples)):
        x = prompt_ids[None, :]
        y = model.generate(
            x,
            max_new_tokens,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            do_sample=True,
            stop_ids=(tokenizer.unk_id, tokenizer.pad_id),
        )
        output = tokenizer.decode(y[0].tolist())
        output = output[len(prompt) :]
        outputs.append(output)
    return outputs


DEFAULT_G_MODEL = {
    "dna_gpt0.1b_h.pth": {
        "path": os.path.join(
            os.path.dirname(__file__), "DNAGPT/checkpoints/dna_gpt0.1b_h.pth"
        ),
        "url": "https://drive.google.com/uc?id=15m6CH3zaMSqflOaf6ec5VPfiulg-Gh0u",
    },
    "dna_gpt0.1b_m": {
        "path": os.path.join(
            os.path.dirname(__file__), "DNAGPT/checkpoints/dna_gpt0.1b_m.pth"
        ),
        "url": "https://drive.google.com/uc?id=1C0BRXfz7RNtCSjSY1dKQeR1yP7I3wTyx",
    },
    "dna_gpt3b_m.pth": {
        "path": os.path.join(
            os.path.dirname(__file__), "DNAGPT/checkpoints/dna_gpt3b_m.pth.pth"
        ),
        "url": "https://drive.google.com/uc?id=1pQ3Ai7C-ObzKkKTRwuf6eshVneKHzYEg",
    },
}


@cli.command()
@click.option("--seed_path", type=str)
@click.option(
    "--num_samples", type=int, default=100, help="number samples of generation"
)
@click.option("--output_path", type=str)
@click.option("--random_state", type=int, default=42)
@click.option("--g_model_name", type=str, default="dna_gpt0.1b_m")
@click.option(
    "--g_weight_path", type=str, default=DEFAULT_G_MODEL["dna_gpt0.1b_m"]["path"]
)
@click.option("--g_max_len", type=int, default=512)
@click.option("--g_temperature", type=float, default=1.0, help="sample temperature")
@click.option("--g_topk", type=int, default=0, help="sample topk")
@click.option("--g_topp", type=float, default=0.95, help="sample topp")
@click.option("--c_model_name_or_path", type=str, default="liminghong/DNABERT-2-117M")
@click.option(
    "--c_lora_model_path",
    type=str,
    default=os.path.join(os.path.dirname(__file__), "adapter_model"),
)
@click.option("--c_max_len", type=int, default=1250)
def generate_sequence(
    seed_path,
    num_samples,
    output_path,
    random_state,
    g_model_name,
    g_weight_path,
    g_max_len,
    g_temperature,
    g_topk,
    g_topp,
    c_model_name_or_path,
    c_lora_model_path,
    c_max_len,
):
    seeds = pd.read_csv(seed_path, sep="\t", index_col=0)
    seed = seeds.sample(n=1, random_state=random_state)

    torch.set_grad_enabled(False)
    seed_all_rng(random_state)

    if g_model_name in DEFAULT_G_MODEL.keys():
        _info = DEFAULT_G_MODEL[g_model_name]
        if g_weight_path == _info["path"] and not os.path.exists(_info["path"]):
            gdown.download(_info["url"], _info["path"])

    prompt = "<R>" + seed["sequence"].item()
    model, tokenizer = get_model(g_model_name)
    model = load_model(model, g_weight_path, device="cuda", dtype=torch.float16)
    texts = generate(
        model,
        tokenizer,
        prompt,
        max_len=min(g_max_len, model.max_len),
        num_samples=num_samples,
        temperature=g_temperature,
        top_k=g_topk,
        top_p=g_topp,
    )

    scores = classify(texts, c_model_name_or_path, c_lora_model_path, c_max_len)
    predict_results = pd.DataFrame({"text": texts, "score": scores})
    predict_results = predict_results.sort_values(by="score", ascending=False)
    predict_results.to_csv(output_path)


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
# Copyright (c) 2023, Tencent Inc. All rights reserved.
# Author: chenchenqin
# Data: 2023/2/2 12:07
import math
import numbers
from typing import Union, List, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
from ..utils import top_k_top_p_filter
_shape_t = Union[int, List[int], torch.Size]


class LayerNorm(nn.Module):
    """LayerNorm with optional bias. See description of nn.LayerNorm"""
    __constants__ = ['normalized_shape', 'eps', 'elementwise_affine']
    normalized_shape: Tuple[int, ...]
    eps: float
    elementwise_affine: bool

    def __init__(self,
                 normalized_shape: _shape_t,
                 eps: float = 1e-5,
                 elementwise_affine: bool = True,
                 bias: bool = True,
                 device=None,
                 dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        super(LayerNorm, self).__init__()
        if isinstance(normalized_shape, numbers.Integral):
            # mypy error: incompatible types in assignment
            normalized_shape = (normalized_shape,)  # type: ignore[assignment]
        self.normalized_shape = tuple(normalized_shape)  # type: ignore[arg-type]
        self.eps = eps
        self.elementwise_affine = elementwise_affine
        if self.elementwise_affine:
            self.weight = Parameter(torch.empty(self.normalized_shape, **factory_kwargs))
            self.bias = Parameter(torch.empty(self.normalized_shape, **factory_kwargs)) if bias else None
        else:
            assert not bias, f"when elementwise_affine = False, bias item is None"
            self.register_parameter('weight', None)
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        if self.elementwise_affine:
            nn.init.ones_(self.weight)
            if self.bias is not None:
                nn.init.zeros_(self.bias)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return F.layer_norm(
            input, self.normalized_shape, self.weight, self.bias, self.eps)

    def extra_repr(self) -> str:
        return '{normalized_shape}, eps={eps}, ' \
               'elementwise_affine={elementwise_affine}'.format(**self.__dict__)


class MultiheadAttention(nn.Module):
    """multi-head attention like torch.nn.MultiheadAttention
    """

    def __init__(self, embedding_dim, num_heads, bias=True, dropout=0.0):
        super().__init__()
        assert embedding_dim % num_heads == 0
        # Q, K, V
        self.c_attn = nn.Linear(embedding_dim, 3 * embedding_dim, bias=bias)
        # output projection
        self.c_proj = nn.Linear(embedding_dim, embedding_dim, bias=bias)
        # regularization
        self.dropout = dropout
        self.attn_dropout = nn.Dropout(self.dropout)
        self.res_dropout = nn.Dropout(self.dropout)
        self.num_heads = num_heads
        self.embedding_dim = embedding_dim

    def forward(self, x, attn_mask=None):
        """
        Args；
            x: tensor[bs, seq_len, embedding_dim]
            attn_mask: mask for masked attention

        Returns:
            tensor[bs, seq_len, embedding_dim]
        """
        B, T, C = x.size()
        # calculate query, key, values for all heads in batch and move head forward to be the batch dim
        q, k, v = self.c_attn(x).split(self.embedding_dim, dim=2)
        k = k.view(B, T, self.num_heads, C // self.num_heads).transpose(1, 2)  # (B, nh, T, hs)
        q = q.view(B, T, self.num_heads, C // self.num_heads).transpose(1, 2)  # (B, nh, T, hs)
        v = v.view(B, T, self.num_heads, C // self.num_heads).transpose(1, 2)  # (B, nh, T, hs)
        # pytorch build-in attention, efficient attention using Flash Attention CUDA kernels
        if hasattr(F, 'scaled_dot_product_attention'):
            assert self.dropout == 0.0, "need dropout=0.0 for now, PyTorch team is working on fix in #92917"
            y = F.scaled_dot_product_attention(q, k, v,
                                               attn_mask=None,
                                               dropout_p=self.dropout,
                                               is_causal=True)
        else:
            # self-attention (B, nh, T, hs) x (B, nh, hs, T) -> (B, nh, T, T)
            att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
            # causal mask to ensure that attention is only applied to the left in the input sequence
            if attn_mask is None:
                attn_mask = torch.tril(torch.ones(T, T, dtype=x.dtype, device=x.device)).view(1, 1, T, T)

            att = att.masked_fill(attn_mask == 0, float('-inf'))
            att = F.softmax(att, dim=-1)
            att = self.attn_dropout(att)
            y = att @ v  # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)

        y = y.transpose(1, 2).contiguous().view(B, T, C)  # re-assemble all head outputs side by side
        y = self.c_proj(y)
        y = self.res_dropout(y)

        return y


class MLP(nn.Module):
    """feed forward"""

    def __init__(self,
                 dim: int,
                 hidden_dim: int,
                 bias: bool = True,
                 dropout: float = 0.0):
        super().__init__()
        self.c_fc = nn.Linear(dim, hidden_dim, bias=bias)
        self.gelu = nn.GELU('tanh')
        self.c_proj = nn.Linear(hidden_dim, dim, bias=bias)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        x = self.c_fc(x)
        x = self.gelu(x)
        x = self.c_proj(x)
        x = self.dropout(x)

        return x


class Block(nn.Module):
    """Transformer block"""

    def __init__(self,
                 dim: int,
                 num_heads: int,
                 dropout: float = 0.0,
                 bias=True):
        super().__init__()
        self.attn = MultiheadAttention(dim, num_heads, dropout=dropout, bias=bias)
        self.mlp = MLP(dim, hidden_dim=4 * dim, bias=bias, dropout=dropout)
        self.ln_1 = LayerNorm(dim, bias=bias)
        self.ln_2 = LayerNorm(dim, bias=bias)

    def forward(self, x):
        x = x + self.attn(self.ln_1(x))
        x = x + self.mlp(self.ln_2(x))

        return x


class GPT(nn.Module):
    CONFIG = {
        # GPT-1
        'openai-gpt': dict(num_layers=12, num_heads=12, embedding_dim=768),  # 117M params
        # GPT-2 configs
        'gpt2': dict(num_layers=12, num_heads=12, embedding_dim=768),  # 124M params
        'gpt2-medium': dict(num_layers=24, num_heads=16, embedding_dim=1024),  # 350M params
        'gpt2-large': dict(num_layers=36, num_heads=20, embedding_dim=1280),  # 774M params
        'gpt2-xl': dict(num_layers=48, num_heads=25, embedding_dim=1600),  # 1558M params
        # Gophers
        'gopher-44m': dict(num_layers=8, num_heads=16, embedding_dim=512),
        'gpt-mini': dict(num_layers=6, num_heads=6, embedding_dim=192),
        'gpt-micro': dict(num_layers=4, num_heads=4, embedding_dim=128),
        'gpt-nano': dict(num_layers=3, num_heads=3, embedding_dim=48)
    }

    def __init__(self,
                 vocab_size,
                 max_len=1024,
                 num_layers=3,
                 num_heads=3,
                 embedding_dim=48,
                 dropout=0.0,
                 bias=True,
                 include_head=True
                 ):
        super().__init__()
        self.vocab_size = vocab_size
        self.num_layers = num_layers
        self.max_len = max_len
        self.num_heads = num_heads
        self.embedding_dim = embedding_dim
        self.dropout = dropout
        self.Block = Block
        self.transformer = nn.ModuleDict(dict(
            wte=nn.Embedding(self.vocab_size, self.embedding_dim),
            wpe=nn.Embedding(self.max_len, self.embedding_dim),
            drop=nn.Dropout(self.dropout),
            h=nn.ModuleList(
                [Block(self.embedding_dim, self.num_heads, dropout=dropout, bias=bias) for _ in
                 range(self.num_layers)]),
            ln_f=LayerNorm(self.embedding_dim, bias=bias),
        ))
        self.lm_head = nn.Linear(self.embedding_dim, self.vocab_size, bias=bias) if include_head else None
        for pn, p in self.named_parameters():
            if pn.endswith('c_proj.weight'):
                torch.nn.init.normal_(p, mean=0.0, std=0.02 / math.sqrt(2 * self.num_layers))
        # report number of parameters (note we don't count the decoder parameters in lm_head)
        n_params = sum(p.numel() for p in self.transformer.parameters())
        print("number of parameters: %.2fM" % (n_params / 1e6,))

    def forward(self, input_ids):
        bs, seq_len = input_ids.shape
        tok_emb = self.transformer.wte(input_ids)
        pos = torch.arange(0, seq_len, dtype=torch.long, device=input_ids.device).unsqueeze(0)  # (1, seq_len)
        pos_emb = self.transformer.wpe(pos)  # position embeddings of shape (1, tseq_len, n_embed)
        emb = tok_emb + pos_emb
        x = self.transformer.drop(emb)
        for block in self.transformer.h:
            x = block(x)
        x = self.transformer.ln_f(x)
        # [bs, seq_len, n_embed]
        if self.lm_head is not None:
            x = self.lm_head(x)

        return x

    @torch.no_grad()
    def generate(self,
                 idx,
                 max_new_tokens,
                 temperature=1.0,
                 do_sample=False,
                 top_k=0,
                 top_p=0.,
                 stop_ids=None):
        """
        Take a conditioning sequence of indices idx (LongTensor of shape (b,t)) and complete
        the sequence max_new_tokens times, feeding the predictions back into the model each time.
        Most likely you'll want to make sure to be in model.eval() mode of operation for this.
        """
        for _ in range(max_new_tokens):
            # if the sequence context is growing too long we must crop it at block_size
            idx_cond = idx if idx.size(1) <= self.max_len else idx[:, -self.max_len:]
            # forward the model to get the logits for the index in the sequence
            logits = self(idx_cond)
            # pluck the logits at the final step and scale by desired temperature
            logits = logits[:, -1, :] / temperature
            # optionally crop the logits to only the top k options
            # if top_k:
            #     v, _ = torch.topk(logits, top_k)
            #     logits[logits < v[:, [-1]]] = -float('Inf')
            logits = top_k_top_p_filter(logits, top_k=top_k, top_p=top_p)
            # apply softmax to convert logits to (normalized) probabilities
            probs = F.softmax(logits, dim=-1)
            # either sample from the distribution or take the most likely element
            if do_sample:
                idx_next = torch.multinomial(probs, num_samples=1)
            else:
                _, idx_next = torch.topk(probs, k=1, dim=-1)

            if stop_ids is not None and idx_next.item() in stop_ids:
                break

            # append sampled index to the running sequence and continue
            idx = torch.cat((idx, idx_next), dim=1)

        return idx